#include <linux/module.h>
#include <linux/cdev.h>
#include <linux/syscalls.h>
#include <linux/uaccess.h>
#include <linux/device.h>
#include <linux/gpio.h>
#include <linux/platform_device.h>
#include <linux/of_address.h>
#include <linux/of_irq.h>
#include <linux/of_gpio.h>
#include <linux/clk.h>
#include <linux/pm_runtime.h>
#include <linux/delay.h>

#include "sensor_phy.h"
#include "sensor_phy_regs.h"

#define BITMASK(num)	((1 << (num)) - 1)


/*
 * Sensor interface muxing
 *
 *			data					clk
 *			sensor_top_ctrl[1:0] 	sensor_top_ctrl[3:2]
 *
 *	GPIO	00						00
 *	LVDS	01						01
 *	HISPI	10						10
 *
 */
#define MUX_SEL_GPIO	0x0
#define MUX_SEL_LVDS	0x5
#define MUX_SEL_HISPI	0xA

#define HISPI_16_BIT_WORDLOCK0 0x7FFFFFFF
#define HISPI_WORDLOCK0(x)	(HISPI_16_BIT_WORDLOCK0 >> (2*(16-(x))))

#define HISPI_16_BIT_WORDLOCK1 0x80000000
#define HISPI_WORDLOCK1(x)	(HISPI_16_BIT_WORDLOCK1 >> (2*(16-(x))))

#define HISPI_16_BIT_WORDLOCK2 0x80000000
#define HISPI_WORDLOCK2(x)	(HISPI_16_BIT_WORDLOCK2 >> (2*(16-(x))))

#define HISPI_16_BIT_SYNCCODE0_00 0x0000FFFF
#define HISPI_SYNCCODE0_00(x)	(HISPI_16_BIT_SYNCCODE0_00 >> (16-(x)))

#define HISPI_16_BIT_SYNCCODE0_01 0x80000000
#define HISPI_SYNCCODE0_01(x)	(HISPI_16_BIT_SYNCCODE0_01 >> (4*(16-(x))))

#define HISPI_SYNCCODE1_00 0x00000000
#define HISPI_SYNCCODE1_01 0x00000000

#define HISPI_16_BIT_SYNCCODE2_00 0x0000FFFF
#define HISPI_SYNCCODE2_00(x)	(HISPI_16_BIT_SYNCCODE2_00 >> (16-(x)))

#define HISPI_16_BIT_SYNCCODE2_01 0xAB000000
#define HISPI_SYNCCODE2_01(x)	(HISPI_16_BIT_SYNCCODE2_01 >> (4*(16-(x))))

#define HISPI_SYNCCODE3_00 0x00000000
#define HISPI_SYNCCODE3_01 0x00000000

#define NUM_CLOCKS 4
//#define MOD_CLOCK_TEST

static bool device_detected = false;
static bool driver_registered = false;

struct sensor_phy {
	u8 enabled_phy;
	u8 enabled_ref_clk;
	u32 sensor_csr_reg_base;
	u32 sensor_csr_reg_size;
	void __iomem *sensor_csr_base;
	u32 sensor_csr_mipi_reg_base;
	u32 sensor_csr_mipi_reg_size;
	void __iomem *sensor_csr_mipi_base;
	u32 mipi_top_reg_base;
	u32 mipi_top_reg_size;
	void __iomem *mipi_top_base;
	struct clk *mipi_cfg_clk;
	u32 mipi_cfg_clk_rate;
	struct clk *av_out0_clk;
	u32 av_out0_clk_rate;
	struct clk *lvds_clk;
	u32 lvds_clk_rate;
	struct clk *isp_pixel_clk;
	u32 isp_pixel_clk_rate;
	struct clk *pll_clk;
	struct clk *xtal_mux_clk;
	struct clk *lvds_mux_clk;

	struct cdev cdev;
	struct class *class;
	struct device* dev;
	dev_t dev_str;
	int sensor_major;
	int reset_gpio;
	int pwdn_gpio;
};

static int sensor_dev_count = 0;

static int sensor_phy_probe(struct platform_device *ofdev);
static int sensor_phy_remove(struct platform_device *ofdev);
static int sensor_phy_suspend(struct device *ofdev);
static int sensor_phy_resume(struct device *ofdev);

static SIMPLE_DEV_PM_OPS(sensor_phy_pm_ops, sensor_phy_suspend, sensor_phy_resume);

#ifdef CONFIG_OF
static const struct of_device_id of_ids[] = {
	{ .compatible = "img,sensor_phy" },
	{ }
};
MODULE_DEVICE_TABLE(of, of_ids);
#endif

static struct platform_driver local_driver = {
    .probe  = sensor_phy_probe,
    .remove  = sensor_phy_remove,
    .driver = { .name = "img,sensor_phy", .owner = THIS_MODULE, .of_match_table = of_match_ptr(of_ids), .pm = &sensor_phy_pm_ops},
};

static inline void sensor_phy_writel(void __iomem *base, unsigned int offset, unsigned int value)
{
	writel(value, base + offset);
}

static inline unsigned int sensor_phy_readl(void __iomem *base, unsigned int offset)
{
	return readl(base + offset);
}

static int sensor_phy_status(struct sensor_phy* phy, struct sensor_phy_status* status)
{
	int ret = 0;
	u32 v;

	v = sensor_phy_readl(phy->sensor_csr_base, SENSOR_CSR_SENSOR_STATUS0_OFFSET);
	status->status0 = (v & SENSOR_CSR_SENSOR_STATUS0_SENSOR_STATUS0_PHY0_MASK) >> SENSOR_CSR_SENSOR_STATUS0_SENSOR_STATUS0_PHY0_SHIFT;
	status->status1 = (v & SENSOR_CSR_SENSOR_STATUS0_SENSOR_STATUS0_PHY1_MASK) >> SENSOR_CSR_SENSOR_STATUS0_SENSOR_STATUS0_PHY1_SHIFT;
	v = sensor_phy_readl(phy->sensor_csr_base, SENSOR_CSR_SENSOR_STATUS1_OFFSET);
	status->status2 = (v & SENSOR_CSR_SENSOR_STATUS1_SENSOR_STATUS0_PHY2_MASK) >> SENSOR_CSR_SENSOR_STATUS1_SENSOR_STATUS0_PHY2_SHIFT;
	status->status3 = (v & SENSOR_CSR_SENSOR_STATUS1_SENSOR_STATUS0_PHY3_MASK) >> SENSOR_CSR_SENSOR_STATUS1_SENSOR_STATUS0_PHY3_SHIFT;

	return ret;
}

static int sensor_phy_reset_gpio(struct sensor_phy* phy, struct sensor_phy_params* params)
{
	int gpio_value = params->reset_gpio;

	gpio_set_value_cansleep(phy->reset_gpio, gpio_value);

	return 0;
}

static int sensor_phy_clk_config(struct sensor_phy* phy, struct sensor_phy_params* params)
{
	int ret = 0;

	if (!phy->av_out0_clk || IS_ERR(phy->av_out0_clk)) {
		pr_err("bad clock\n");
		return -EFAULT;
	}

	/* Enable/disable av_out0_clk - the reference input clock for all sensors */
	if(params->clk_en && !phy->enabled_ref_clk) {
		if (clk_get_rate(phy->av_out0_clk) != params->clk_freq)
			phy->av_out0_clk_rate = params->clk_freq;
		ret = clk_set_rate(phy->av_out0_clk, phy->av_out0_clk_rate);
		if(ret)
			goto out_err_clk_set;
		clk_prepare_enable(phy->av_out0_clk);
		phy->enabled_ref_clk = 1;
	} else if(!params->clk_en && phy->enabled_ref_clk) {
		clk_disable_unprepare(phy->av_out0_clk);
		phy->enabled_ref_clk = 0;
	}

	return 0;

out_err_clk_set:
	return ret;
}

static int sensor_phy_lvds_clk_config(struct sensor_phy* phy, __u8 enable)
{
	unsigned long rate;
	int ret;
	struct clk *clock = phy->pll_clk;

	if(enable) {

		ret = clk_set_rate(phy->lvds_clk, phy->lvds_clk_rate);
		if(ret) {
			printk(KERN_ERR "sensor_phy_lvds_clk_config - error setting lvds_clk\n");
			return ret;
		}

		clk_set_parent(phy->lvds_mux_clk, phy->lvds_clk);

		/* For HiSPI1, num_phy=4 */
		/* isp_pixel_clk_rate = ((lvds_clk_rate * 2 * 4 * num_phy) / pixel_size) */
		/* Let pll_rate = isp_pixel_clk_rate */
		ret = clk_set_rate(clock, phy->isp_pixel_clk_rate);
		if(ret) {
			printk(KERN_ERR "sensor_phy_lvds_clk_config - failed to set pll_clk to %u\n",
					phy->isp_pixel_clk_rate);
			return ret;
		}

		rate = clk_get_rate(clock);
		if (rate != phy->isp_pixel_clk_rate) {
			printk(KERN_ERR "req rate of %u not achieved. rate = %lu\n",
					phy->isp_pixel_clk_rate, rate);
			return ret;
		}

		ret = clk_set_rate(phy->isp_pixel_clk, phy->isp_pixel_clk_rate);
		if(ret) {
			printk(KERN_ERR "sensor_phy_lvds_clk_config - error setting isp_pixel_clk\n");
			return ret;
		}
		clk_prepare_enable(clock);
		clk_prepare_enable(phy->isp_pixel_clk);
	} else {
		/* Re-parent the PLL to XTAL */
		clk_set_parent(phy->lvds_mux_clk, phy->xtal_mux_clk);

		if(!IS_ERR(phy->isp_pixel_clk)) {
			clk_disable_unprepare(phy->isp_pixel_clk);
		}
	}

	return 0;
}

static int sensor_phy_hispi_mux_config(struct sensor_phy* phy, struct sensor_phy_params* params)
{
	void __iomem *base;
	unsigned int regval = 0;

	base = phy->sensor_csr_base;

	/* SENSOR_TOP_CTRL - MUX select */
	printk(KERN_DEBUG "%s HiSPI MUX\n", params->enable_mux ? "Select" : "Deselect");
	regval = params->enable_mux ? MUX_SEL_HISPI : MUX_SEL_GPIO;
	sensor_phy_writel(base, SENSOR_CSR_SENSOR_TOP_CTRL_OFFSET, regval);

	return 0;
}

static int sensor_phy_config(struct sensor_phy* phy, struct sensor_phy_params* params)
{
	void __iomem *base, *base_mipi, *base_mipi_top;
	unsigned int regval = 0;
	u8 mipi_lanes_en = 0;
	int ret = 0;

	base = phy->sensor_csr_base;
	base_mipi = phy->sensor_csr_mipi_base;
	base_mipi_top = phy->mipi_top_base;

	switch(params->type)
	{
		case SENSOR_IF_HISPI:

			if(params->alt_mode) {
				/*
				 * When using HiSPI1:
				 * Configure isp_pixel_clk, with lvds_clk as source
				 */
				if(sensor_phy_lvds_clk_config(phy, params->enable)) {
					printk(KERN_ERR "sensor_phy_config - error configuring lvds_clk\n");
					break;
				}
			}

			/* SENSOR_CTRL (Disable HiSPI) */
			regval = (0x3 << SENSOR_CSR_SENSOR_CTRL_FIFO_AEMPTY_THRESH_SHIFT |
					  0x0 << SENSOR_CSR_SENSOR_CTRL_FIFO_AFULL_THRESH_SHIFT |
					  0x0 << SENSOR_CSR_SENSOR_CTRL_ENABLE_HISPI_SHIFT);
			sensor_phy_writel(base, SENSOR_CSR_SENSOR_CTRL_OFFSET, regval);

			if(!params->enable)
				break;

			/* WORDLOCK0 */
			regval = HISPI_WORDLOCK0(params->pixel_res);
			sensor_phy_writel(base, SENSOR_CSR_WORDLOCK0_OFFSET, regval);

			/* WORDLOCK1 */
			regval = HISPI_WORDLOCK1(params->pixel_res);
			sensor_phy_writel(base, SENSOR_CSR_WORDLOCK1_OFFSET, regval);

			/* WORDLOCK2 */
			regval = HISPI_WORDLOCK2(params->pixel_res);
			sensor_phy_writel(base, SENSOR_CSR_WORDLOCK2_OFFSET, regval);

			/* SYNCCODE0_00 && SYNCCODE0_01 */
			regval = HISPI_SYNCCODE0_00(params->pixel_res);
			sensor_phy_writel(base, SENSOR_CSR_SYNC_CODE0_00_OFFSET, regval);
			regval = HISPI_SYNCCODE0_01(params->pixel_res);
			sensor_phy_writel(base, SENSOR_CSR_SYNC_CODE0_01_OFFSET, regval);

			/* SYNCCODE1_00 && SYNCCODE1_01 */
			regval = HISPI_SYNCCODE1_00;
			sensor_phy_writel(base, SENSOR_CSR_SYNC_CODE1_00_OFFSET, regval);
			regval = HISPI_SYNCCODE1_01;
			sensor_phy_writel(base, SENSOR_CSR_SYNC_CODE1_01_OFFSET, regval);

			/* SYNCCODE2_00 && SYNCCODE2_01 */
			regval = HISPI_SYNCCODE2_00(params->pixel_res);
			sensor_phy_writel(base, SENSOR_CSR_SYNC_CODE2_00_OFFSET, regval);
			regval = HISPI_SYNCCODE2_01(params->pixel_res);
			sensor_phy_writel(base, SENSOR_CSR_SYNC_CODE2_01_OFFSET, regval);

			/* SYNCCODE3_00 && SYNCCODE3_01 */
			regval = HISPI_SYNCCODE3_00;
			sensor_phy_writel(base, SENSOR_CSR_SYNC_CODE3_00_OFFSET, regval);
			regval = HISPI_SYNCCODE3_01;
			sensor_phy_writel(base, SENSOR_CSR_SYNC_CODE3_01_OFFSET, regval);

			/* ACTIVE_PIXEL_SIZE */
			regval = params->frame_width;
			sensor_phy_writel(base, SENSOR_CSR_ACTIVE_PIXEL_SIZE_OFFSET, regval);

			/* SENSOR_CTRL (Enable HiSPI and program config) */
			regval = 0;
			regval = (params->alt_mode << SENSOR_CSR_SENSOR_CTRL_EXT_CLK_SHIFT |
					  params->alt_mode << SENSOR_CSR_SENSOR_CTRL_WORD_LOCK_CTRL_SHIFT |
					  0x3 << SENSOR_CSR_SENSOR_CTRL_FIFO_AEMPTY_THRESH_SHIFT |
					  0x0 << SENSOR_CSR_SENSOR_CTRL_FIFO_AFULL_THRESH_SHIFT |
					  0x0 << SENSOR_CSR_SENSOR_CTRL_TWOPIXEL_MODE_SHIFT |
					  params->filler_en << SENSOR_CSR_SENSOR_CTRL_FILLERCODE_EN_SHIFT |
					  0x1 << SENSOR_CSR_SENSOR_CTRL_ENABLE_HISPI_SHIFT |
					  (params->num_phys - 1) << SENSOR_CSR_SENSOR_CTRL_NUM_PHY_SHIFT |
					  !params->lsb_first << SENSOR_CSR_SENSOR_CTRL_DATA_MSB_PHY_SHIFT |
					  0x0 << SENSOR_CSR_SENSOR_CTRL_VSYNC_COND_SHIFT |
					  0x0 << SENSOR_CSR_SENSOR_CTRL_HSYNC_COND_SHIFT |
					  ((params->pixel_res-10)/2) << SENSOR_CSR_SENSOR_CTRL_WORD_SIZE_SHIFT);
			sensor_phy_writel(base, SENSOR_CSR_SENSOR_CTRL_OFFSET, regval);
			break;

		case SENSOR_IF_LVDS:

			/* SENSOR_TOP_CTRL - MUX select */
			regval = params->enable ? MUX_SEL_LVDS: MUX_SEL_GPIO;
			sensor_phy_writel(base, SENSOR_CSR_SENSOR_TOP_CTRL_OFFSET, regval);
			break;

		case SENSOR_IF_GPIO:

			/* Pad muxing and config determined by device-tree config */

			/* SENSOR_TOP_CTRL - MUX select */
			regval = MUX_SEL_GPIO;
			sensor_phy_writel(base, SENSOR_CSR_SENSOR_TOP_CTRL_OFFSET, regval);
			break;

		case SENSOR_IF_MIPI:

			regval = sensor_phy_readl(base_mipi, SENSOR_CSR_PHY0_TST_CTRL0_OFFSET);
			regval |= (0x1 << SENSOR_CSR_PHY0_TST_CTRL0_PHY0_TESTCLR_SHIFT);
			sensor_phy_writel(base_mipi, SENSOR_CSR_PHY0_TST_CTRL0_OFFSET, regval);
			ndelay(500);

			if (params->enable) {
				ret = clk_set_rate(phy->mipi_cfg_clk, phy->mipi_cfg_clk_rate);
				if(ret) {
					printk(KERN_ERR "sensor_phy_config - error setting mipi_cfg_clk\n");
					break;
				}
				clk_prepare_enable(phy->mipi_cfg_clk);

				regval |= (0x0 << SENSOR_CSR_PHY0_TST_CTRL0_PHY0_TESTCLR_SHIFT);
				sensor_phy_writel(base_mipi, SENSOR_CSR_PHY0_TST_CTRL0_OFFSET, regval);
				ndelay(200);
				regval = sensor_phy_readl(base_mipi, SENSOR_CSR_PHY0_TST_CTRL1_OFFSET);
				regval |= (0x44 << SENSOR_CSR_PHY0_TST_CTRL1_PHY0_TESTDIN_SHIFT);
				sensor_phy_writel(base_mipi, SENSOR_CSR_PHY0_TST_CTRL1_OFFSET, regval);
				regval = sensor_phy_readl(base_mipi, SENSOR_CSR_PHY0_TST_CTRL0_OFFSET);
				regval |= (0x1 << SENSOR_CSR_PHY0_TST_CTRL0_PHY0_TESTCLK_SHIFT);
				sensor_phy_writel(base_mipi, SENSOR_CSR_PHY0_TST_CTRL0_OFFSET, regval);
				ndelay(500);
				regval = sensor_phy_readl(base_mipi, SENSOR_CSR_PHY0_TST_CTRL1_OFFSET);
				regval |= (0x1 << SENSOR_CSR_PHY0_TST_CTRL1_PHY0_TESTEN_SHIFT);
				sensor_phy_writel(base_mipi, SENSOR_CSR_PHY0_TST_CTRL1_OFFSET, regval);
				regval |= (0x0 << SENSOR_CSR_PHY0_TST_CTRL0_PHY0_TESTCLK_SHIFT);
				sensor_phy_writel(base_mipi, SENSOR_CSR_PHY0_TST_CTRL0_OFFSET, regval);
				ndelay(720);
				regval = sensor_phy_readl(base_mipi, SENSOR_CSR_PHY0_TST_CTRL1_OFFSET);
				regval |= (0x54 << SENSOR_CSR_PHY0_TST_CTRL1_PHY0_TESTDIN_SHIFT);
				sensor_phy_writel(base_mipi, SENSOR_CSR_PHY0_TST_CTRL1_OFFSET, regval);
				ndelay(50);
				regval = sensor_phy_readl(base_mipi, SENSOR_CSR_PHY0_TST_CTRL1_OFFSET);
				regval |= (0x0 << SENSOR_CSR_PHY0_TST_CTRL1_PHY0_TESTEN_SHIFT);
				sensor_phy_writel(base_mipi, SENSOR_CSR_PHY0_TST_CTRL1_OFFSET, regval);
				ndelay(160);
				regval = sensor_phy_readl(base_mipi, SENSOR_CSR_PHY0_TST_CTRL0_OFFSET);
				regval |= (0x1 << SENSOR_CSR_PHY0_TST_CTRL0_PHY0_TESTCLK_SHIFT);
				sensor_phy_writel(base_mipi, SENSOR_CSR_PHY0_TST_CTRL0_OFFSET, regval);
				ndelay(180);
			}

			mipi_lanes_en = params->enable ? BITMASK(params->mipi_lanes) : 0;
			regval = (mipi_lanes_en << MIPI0_PHY_CTRL_CR_MIPI0_PHY_ENABLE_RX_LANE_0_SHIFT |
					  params->enable << MIPI0_PHY_CTRL_CR_MIPI0_PHY_ENABLECLK_SHIFT |
					  params->enable << MIPI0_PHY_CTRL_CR_MIPI0_PHY_SHUTDOWNZ_SHIFT |
					  params->enable << MIPI0_PHY_CTRL_CR_MIPI0_PHY_RSTZ_SHIFT);

			sensor_phy_writel(base_mipi_top, MIPI0_PHY_CTRL_OFFSET, regval);

			if (!params->enable && !IS_ERR(phy->mipi_cfg_clk)) {
				clk_disable_unprepare(phy->mipi_cfg_clk);
			}
			break;

		default:
			printk(KERN_ERR "sensor_phy_config_ctl - unknown sensor type\n");
			ret = -EFAULT;
	}

	if (ret == 0) {
		if(params->enable)
			phy->enabled_phy |= params->type;
		else
			phy->enabled_phy &= ~params->type;
	}

	return ret;
}

static long sensor_phy_ioctl(struct file * filp, unsigned int ioctl, unsigned long arg)
{
	void __user *argp = (void __user *)arg;
	struct sensor_phy *phy = (struct sensor_phy*)filp->private_data;
	union {
		struct sensor_phy_params params;
		struct sensor_phy_status status;
	} args;
	int ret = 0;

	switch (ioctl)
	{
		case IOC_SENSOR_PHY_CONFIG:
			if (copy_from_user(&args.params, argp, sizeof(args.params)))
				return -EFAULT;
			ret = sensor_phy_config(phy, &args.params);
			break;
		case IOC_SENSOR_PHY_CLK_CONFIG:
			if (copy_from_user(&args.params, argp, sizeof(args.params)))
				return -EFAULT;
			ret = sensor_phy_clk_config(phy, &args.params);
			break;
		case IOC_SENSOR_PHY_STATUS:
			ret = sensor_phy_status(phy, &args.status);
			if (copy_to_user(argp, &args.status, sizeof(args.status)))
				return -EFAULT;
			break;
		case IOC_SENSOR_PHY_RESET_GPIO:
			if (copy_from_user(&args.params, argp, sizeof(args.params)))
				return -EFAULT;
			ret = sensor_phy_reset_gpio(phy, &args.params);
			break;
		case IOC_SENSOR_PHY_MUX_CONFIG:
			if (copy_from_user(&args.params, argp, sizeof(args.params)))
				return -EFAULT;
			ret = sensor_phy_hispi_mux_config(phy, &args.params);
			break;
		default:
			printk(KERN_ERR "sensor_phy_ioctl - unknown ioctl value\n");
			ret = -ENOTTY;
	}

	return ret;
}

static int sensor_phy_open(struct inode* inode, struct file* filp)
{
	struct sensor_phy* phy;

	phy = container_of(inode->i_cdev, struct sensor_phy, cdev);
	if (!phy)
	{
		printk(KERN_ERR "%s: Failed to get pointer to phy structure\n", DEV_STR);
		return -EINVAL;
	}

	filp->private_data = (void*)phy;

	return 0;
}

static int sensor_phy_release(struct inode* inode, struct file* filp)
{
	return 0;
}

static struct file_operations sensor_fops =
{
	.owner =  THIS_MODULE,
	.unlocked_ioctl = sensor_phy_ioctl,
	.open = sensor_phy_open,
	.release = sensor_phy_release,
};

static int sensor_phy_remove(struct platform_device *ofdev)
{
	struct sensor_phy *sensor_dev = platform_get_drvdata(ofdev);
	struct sensor_phy_params params = {0};

	/* Disable any active sensor PHYs */
	params.type = SENSOR_IF_HISPI;
	while (sensor_dev->enabled_phy != 0) {
		if (sensor_dev->enabled_phy & params.type)
			sensor_phy_config(sensor_dev, &params);
		params.type <<= 1;
	}

	/* Disable sensor reference clock */
	sensor_phy_clk_config(sensor_dev, &params);

	/* Free reset and powerdown gpio */
	gpio_free(sensor_dev->reset_gpio);
	gpio_free(sensor_dev->pwdn_gpio);

	iounmap(sensor_dev->sensor_csr_base);
	iounmap(sensor_dev->sensor_csr_mipi_base);
	iounmap(sensor_dev->mipi_top_base);

	device_destroy(sensor_dev->class, MKDEV(sensor_dev->sensor_major, 0));
	class_destroy(sensor_dev->class);
	unregister_chrdev_region(sensor_dev->dev_str, 1);

	pm_runtime_disable(&ofdev->dev);

	--sensor_dev_count;

	return 0;
}

#ifdef MOD_CLOCK_TEST
void sensor_phy_clk_test(struct sensor_phy *sensor_dev, int dev_count)
{
	int ret;

	printk(KERN_DEBUG "Clocks for sensor_phy%d\n", dev_count);

	/* MIPI */
	printk(KERN_DEBUG "mipi_cfg_clk_rate (dts) = %d\n", sensor_dev->mipi_cfg_clk_rate);
	printk(KERN_DEBUG "mipi_cfg_clk_rate = %ld\n", clk_get_rate(sensor_dev->mipi_cfg_clk));
	ret = clk_set_rate(sensor_dev->mipi_cfg_clk, sensor_dev->mipi_cfg_clk_rate);
	printk(KERN_DEBUG "mipi_cfg_clk = %ld, ret %d\n", clk_get_rate(sensor_dev->mipi_cfg_clk), ret);

	/* av_out0_clk */
	printk(KERN_DEBUG "av_out0_clk (dts) = %d\n", sensor_dev->av_out0_clk_rate);
	printk(KERN_DEBUG "av_out0_clk = %ld\n", clk_get_rate(sensor_dev->av_out0_clk));
	ret = clk_set_rate(sensor_dev->av_out0_clk, sensor_dev->av_out0_clk_rate);
	printk(KERN_DEBUG "av_out0_clk = %ld, ret %d\n", clk_get_rate(sensor_dev->av_out0_clk), ret);

	if(dev_count == 1) {
		/* lvds_clk */
		printk(KERN_DEBUG "lvds_clk (dts) = %d\n", sensor_dev->lvds_clk_rate);
		printk(KERN_DEBUG "lvds_clk = %ld\n", clk_get_rate(sensor_dev->lvds_clk));
		ret = clk_set_rate(sensor_dev->lvds_clk, sensor_dev->lvds_clk_rate);
		printk(KERN_DEBUG "lvds_clk = %ld, ret %d\n", clk_get_rate(sensor_dev->lvds_clk), ret);

		/* isp_pixel_clk */
		printk(KERN_DEBUG "isp_pixel_clk (dts) = %d\n", sensor_dev->isp_pixel_clk_rate);
		printk(KERN_DEBUG "isp_pixel_clk = %ld\n", clk_get_rate(sensor_dev->isp_pixel_clk));
		ret = clk_set_rate(sensor_dev->isp_pixel_clk, sensor_dev->isp_pixel_clk_rate);
		printk(KERN_DEBUG "isp_pixel_clk = %ld, ret %d\n", clk_get_rate(sensor_dev->isp_pixel_clk), ret);
	}
}
#endif

static int sensor_phy_probe(struct platform_device *ofdev)
{
	int ret, index = 0;
	const __be32 *cur;
	struct property	*prop;
	u32 rate, rates[NUM_CLOCKS];
	struct resource *res;
	struct sensor_phy *sensor_dev;
	char dev_name[20];
	char class_name[20];
	enum of_gpio_flags gpio_flags;

	sensor_dev = devm_kzalloc(&ofdev->dev, sizeof(*sensor_dev), GFP_KERNEL);
	if (!sensor_dev)
		return -ENOMEM;

	res = platform_get_resource(ofdev, IORESOURCE_MEM, 0);
	sensor_dev->sensor_csr_base = devm_ioremap_resource(&ofdev->dev, res);
	if(IS_ERR(sensor_dev->sensor_csr_base))
		return PTR_ERR(sensor_dev->sensor_csr_base);

	sensor_dev->sensor_csr_reg_base = res->start;
	sensor_dev->sensor_csr_reg_size = res->end - res->start + 1;

	printk(KERN_DEBUG "sensor_csr_reg_base 0x%x sensor_csr_reg_size %d\n",
		   sensor_dev->sensor_csr_reg_base, sensor_dev->sensor_csr_reg_size);

	res = platform_get_resource(ofdev, IORESOURCE_MEM, 1);
	sensor_dev->sensor_csr_mipi_base = devm_ioremap_resource(&ofdev->dev, res);
	if(IS_ERR(sensor_dev->sensor_csr_mipi_base))
		return PTR_ERR(sensor_dev->sensor_csr_mipi_base);

	sensor_dev->sensor_csr_mipi_reg_base = res->start;
	sensor_dev->sensor_csr_mipi_reg_size = res->end - res->start + 1;

	printk(KERN_DEBUG "sensor_csr_mipi_reg_base 0x%x sensor_csr_mipi_reg_size %d\n",
		   sensor_dev->sensor_csr_mipi_reg_base, sensor_dev->sensor_csr_mipi_reg_size);

	res = platform_get_resource(ofdev, IORESOURCE_MEM, 2);
	sensor_dev->mipi_top_base = devm_ioremap_resource(&ofdev->dev, res);
	if(IS_ERR(sensor_dev->mipi_top_base))
		return PTR_ERR(sensor_dev->mipi_top_base);

	sensor_dev->mipi_top_reg_base = res->start;
	sensor_dev->mipi_top_reg_size = res->end - res->start + 1;

	printk(KERN_DEBUG "mipi_phy_reg_base 0x%x mipi_phy_reg_size %d\n",
		   sensor_dev->mipi_top_reg_base, sensor_dev->mipi_top_reg_size);

	/* Get all the clock rates */
	of_property_for_each_u32(ofdev->dev.of_node, "clock-frequency", prop, cur, rate) {
		/* BUG: If more clocks are specified, rates[] is overflowed */
			rates[index++] = rate;
	}

	if(index != NUM_CLOCKS) {
		printk(KERN_DEBUG "Incorrect number (%d) of clocks in DTS!\n", index);
		return -EFAULT;
	}

	/* Get the mipi_cfg_clk clock */
	sensor_dev->mipi_cfg_clk = devm_clk_get(&ofdev->dev, "mipi_rx_cfg");
	if (IS_ERR(sensor_dev->mipi_cfg_clk))
		return PTR_ERR(sensor_dev->mipi_cfg_clk);

	sensor_dev->mipi_cfg_clk_rate = rates[0];

	/* Get the av_out0_clk clock */
	sensor_dev->av_out0_clk = devm_clk_get(&ofdev->dev, "av_out0_clk");
	if (IS_ERR(sensor_dev->av_out0_clk))
		return PTR_ERR(sensor_dev->av_out0_clk);

	sensor_dev->av_out0_clk_rate = rates[1];

	/* Get the lvds_clk clock */
	sensor_dev->lvds_clk = devm_clk_get(&ofdev->dev, "lvds_clk");
	if (IS_ERR(sensor_dev->lvds_clk))
		printk(KERN_DEBUG "Warning: no lvds_clk for this sensor_phy\n");
	else
		sensor_dev->lvds_clk_rate = rates[2];

	/* Get the isp_pixel_clk clock */
	sensor_dev->isp_pixel_clk = devm_clk_get(&ofdev->dev, "isp_pixel_clk");
	if (IS_ERR(sensor_dev->isp_pixel_clk))
		printk(KERN_DEBUG "Warning: no isp_pixel_clk for this sensor_phy\n");
	else
		sensor_dev->isp_pixel_clk_rate = rates[3];

	/* Get the source pll */
	sensor_dev->pll_clk = devm_clk_get(&ofdev->dev, "pll");
	if (IS_ERR(sensor_dev->pll_clk))
		printk(KERN_DEBUG "Warning: no source PLL for this sensor_phy\n");

	/* Get the xtal_mux */
	sensor_dev->xtal_mux_clk = devm_clk_get(&ofdev->dev, "xtal_mux");
	if (IS_ERR(sensor_dev->xtal_mux_clk))
		printk(KERN_DEBUG "Warning: no xtal_mux for this sensor_phy\n");

	/* Get the lvds_mux */
	sensor_dev->lvds_mux_clk = devm_clk_get(&ofdev->dev, "lvds_mux");
	if (IS_ERR(sensor_dev->lvds_mux_clk))
		printk(KERN_DEBUG "Warning: no lvds_mux for this sensor_phy\n");

	/* Get the reset GPIO */
	sensor_dev->reset_gpio = of_get_gpio_flags(ofdev->dev.of_node, 0, &gpio_flags);
	if(!gpio_is_valid(sensor_dev->reset_gpio)) {
		printk(KERN_DEBUG "Warning: invalid reset_gpio\n");
		if(sensor_dev->reset_gpio != -EPROBE_DEFER) {
			printk(KERN_DEBUG "Warning: gpios parameter missing or invalid\n");
		}
	} else {
		/* Request a GPIO*/
		ret = devm_gpio_request_one(&ofdev->dev, sensor_dev->reset_gpio,
					    GPIOF_DIR_OUT | GPIOF_INIT_LOW, "sensor_reset");
		if (ret < 0) {
			printk(KERN_DEBUG "Unable to request GPIO\n");
		}
	}

	if (of_gpio_count(ofdev->dev.of_node) > 1) {
		/* Get the powerdown GPIO */
		sensor_dev->pwdn_gpio = of_get_gpio_flags(ofdev->dev.of_node, 1, &gpio_flags);
		if(!gpio_is_valid(sensor_dev->pwdn_gpio)) {
			printk(KERN_DEBUG "Warning: invalid pwdn_gpio\n");
			if(sensor_dev->pwdn_gpio != -EPROBE_DEFER) {
				printk(KERN_DEBUG "Warning: gpios parameter missing or invalid\n");
			}
		} else {
			/* Request a GPIO*/
			ret = devm_gpio_request_one(&ofdev->dev, sensor_dev->pwdn_gpio,
						    GPIOF_DIR_OUT | GPIOF_INIT_HIGH, "sensor_pwdn");
			if (ret < 0) {
				printk(KERN_DEBUG "Unable to request GPIO\n");
			}
		}
	}

	device_detected = true;
	pm_runtime_enable(&ofdev->dev);

	sprintf(dev_name, "%s%d", DEV_STR, sensor_dev_count);
	sprintf(class_name, "%s%d", DEV_CLASS_STR, sensor_dev_count);

	ret = alloc_chrdev_region(&sensor_dev->dev_str, 0, 1, dev_name);
	sensor_dev->sensor_major = MAJOR(sensor_dev->dev_str);
	printk(KERN_DEBUG "%s: chrdev ret=%d, MAJ=%d\n", dev_name, ret, sensor_dev->sensor_major);
	if (ret < 0)
	{
		printk(KERN_ERR "%s: Failed to alloc chrdev region\n", dev_name);
		return ret;
	}
	printk(KERN_DEBUG "%s: alloc_chrdev_region ret %d, assigned MAJ=%d", dev_name, ret, sensor_dev->sensor_major);

	sensor_dev->class = class_create(THIS_MODULE, class_name);
	if (IS_ERR(sensor_dev->class))
	{
		ret = PTR_ERR(sensor_dev->class);
		printk(KERN_ERR "%s: Failed to create device class\n", class_name);
		goto out_err_free_chrdev_region;
	}

	cdev_init(&sensor_dev->cdev, &sensor_fops);
	sensor_dev->cdev.owner = THIS_MODULE;
	sensor_dev->cdev.ops = &sensor_fops;
	ret = cdev_add(&sensor_dev->cdev, sensor_dev->dev_str, 1);

	sensor_dev->dev = device_create(sensor_dev->class, NULL, MKDEV(sensor_dev->sensor_major, 0), NULL, class_name);
	if (IS_ERR(sensor_dev->dev))
	{
		ret = PTR_ERR(sensor_dev->dev);
		printk(KERN_ERR "%s: Failed to create device node\n", dev_name);
		goto out_err_free_class;
	}

	if (ret)
	{
		printk(KERN_ERR "%s: Failed to register cdev, MAJ%d, ret%d\n", dev_name, sensor_dev->sensor_major, ret);
		goto out_err_free_chrdev_region;
	}
	else
	{
		printk(KERN_DEBUG "%s: Successfully registered cdev MAJ %d, MIN 0\n", dev_name, sensor_dev->sensor_major);
	}

#ifdef MOD_CLOCK_TEST
	sensor_phy_clk_test(sensor_dev, sensor_dev_count);
#endif

	platform_set_drvdata(ofdev, sensor_dev);
	sensor_dev_count++;

    return 0;

out_err_free_class:
   	class_destroy(sensor_dev->class);
out_err_free_chrdev_region:
   	unregister_chrdev_region(sensor_dev->dev_str, 1);
    	return ret;
}

static int sensor_phy_register(void)
{
	int ret;

    if ( driver_registered )
    {
        printk(KERN_ERR "Sensor device already detected\n");
        return -EINVAL;
    }

    driver_registered = true;

	ret = platform_driver_register(&local_driver);
	if (ret != 0) {
		ret = -ENODEV;
		printk(KERN_ERR "platform_driver_register failed\n");
		goto failure_register;
	}

	if (device_detected != true) {
		printk(KERN_ERR "device_detect failed\n");
		ret = -ENODEV;
		goto failure_detect;
	}

	return 0;

failure_detect:
	platform_driver_unregister(&local_driver);
failure_register:
	driver_registered = false;
	return ret;
}

static int sensor_phy_deregister(void)
{
	printk(KERN_DEBUG "Unregister sensor PHY driver\n");

	if (driver_registered)
		platform_driver_unregister(&local_driver);
	driver_registered = false;

	return 0;
}

static int sensor_phy_suspend(struct device *ofdev)
{
	int result = 0;

	return result;
}

static int sensor_phy_resume(struct device *ofdev)
{
	int result = 0;

	return result;
}

static int __init sensor_phy_init(void)
{
    int ret = 0;
    printk(KERN_DEBUG "driver initialisation\n");

    ret = sensor_phy_register();
    if (ret)
    {
        printk(KERN_ERR "Failed to register the device driver!\n");
        return ret;
    }

    return ret;
}

static void __exit sensor_phy_exit(void)
{
	sensor_phy_deregister();
}

MODULE_AUTHOR("Patrick Conway <paddy.conway@imgtec.com>");
MODULE_LICENSE("GPL");

module_init(sensor_phy_init);
module_exit(sensor_phy_exit);
